from typing import Type, Any, List
from seatools.ioc.config import cfg
from seatools.ioc.utils.value_utils import convert, parse_express_value


class Environment:

    def __init__(self):
        self._active_profiles = None

    def get_property(self, name: str, cls: Type = None) -> Any:
        data = parse_express_value(data=cfg(), express=name)
        if data is None:
            if cls:
                return cls()
            return None
        if cls:
            return convert(data, cls)
        return data

    def get_active_profiles(self) -> List[str]:
        if not self._active_profiles:
            active_profiles = self.get_property("seatools.profiles.active")
            if not active_profiles:
                self._active_profiles = []
            else:
                self._active_profiles = [active_profile.strip() for active_profile in active_profiles.split(',')]
        return self._active_profiles



from typing import Any

from seatools.ioc.aop.aspect import AbstractAspect
from seatools.ioc.aop.matcher import AspectPointExpressionMather
from seatools.ioc.aop.point import JoinPoint


class SelectorAspect(AbstractAspect):

    def __init__(self, aspect: AbstractAspect):
        self.aspect = aspect
        self.mather = AspectPointExpressionMather(self.aspect.pointcut)
        self._cache = {}

    @property
    def pointcut(self) -> str:
        return self.aspect.pointcut

    def _match(self, point: JoinPoint):
        match = self._cache.get(point.path)
        if match is not None:
            return match
        match = self.mather.match(point.path)
        self._cache[point.path] = match
        return match

    def before(self, point: JoinPoint, **kwargs) -> None:
        if self._match(point):
            return self.aspect.before(point, **kwargs)
        return super().before(point, **kwargs)

    def around(self, point: JoinPoint, **kwargs) -> Any:
        if self._match(point):
            return self.aspect.around(point)
        return super().around(point, **kwargs)

    def after(self, point: JoinPoint, **kwargs) -> None:
        if self._match(point):
            return self.aspect.after(point, **kwargs)
        return super().after(point, **kwargs)

    def after_returning(self, point: JoinPoint, return_value: Any, **kwargs) -> None:
        if self._match(point):
            return self.aspect.after_returning(point, return_value, **kwargs)
        return super().after_returning(point, return_value, **kwargs)

    def after_exception(self, point: JoinPoint, ex: Exception, **kwargs) -> None:
        if self._match(point):
            return self.aspect.after_exception(point, ex, **kwargs)
        return super().after_exception(point, ex, **kwargs)



import abc
from typing import Union, Type, Callable, Any


class BeanFactory(abc.ABC):
    """Abstract class for Bean Factory"""

    def __init__(self, **kwargs):
        pass

    @abc.abstractmethod
    def get_bean(self, name: str = None, required_type: Union[Type, Callable] = None):
        """Method to get a bean."""
        raise NotImplementedError

    def get_beans(self, required_type: Type):
        """Method to get a list of beans of a specific type."""
        raise NotImplementedError

    @abc.abstractmethod
    def register_bean(self, name: str, cls, primary: bool = False, order: int = 0, **kwargs) -> Any:
        """Method to register a bean."""
        raise NotImplementedError

    @abc.abstractmethod
    def init(self, **kwargs):
        """Initialization method for creating beans.
        Note:
             Note: This method may be executed repeatedly due to the Python module loading mechanism.
             In order to support pre dependency handling of IOC related operations executed during module loading.
        """
        raise NotImplementedError



import importlib.util
import inspect
from typing import Union, List, Type
from pydantic import BaseModel

from .base import DataFileLoader, DataFileExtractor


class PyDataFileLoader(DataFileLoader):

    def _convert_config_data_to_python(self,
                                       config_data: Union[str, bytes],
                                       **kwargs) -> Union[dict, tuple, list]:
        return eval(config_data)

    def load_file(self,
                  file_path: str,
                  encoding: str = 'utf-8',
                  modelclass: Union[Type[dict], Type[BaseModel]] = dict,
                  **kwargs) -> Union[dict, list, BaseModel]:
        module_name = file_path.split('.')[0].replace('\\', '/').split('/')[-1]
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        members = inspect.getmembers(module)

        data = {}

        for name, obj in members:
            if not inspect.isbuiltin(obj) and not inspect.ismodule(obj) and not name.startswith(
                '__') and not name.endswith('__'):
                data[name] = obj

        return self._convert_item_to_model(data, modelclass)


class PyDataFileExtractor(DataFileExtractor):

    def _convert_data_to_str(self, data: Union[dict, List[dict]], **kwargs) -> str:
        raise NotImplementedError('不支持当前类型文件提取')



import re
from pathlib import Path

from seatools.files import AutoDataFileLoader
from seatools.utils import dict_utils
from seatools.ioc.utils import value_utils
from typing import Optional, Dict, Union
from loguru import logger
import os
import types


class _Properties:
    cfg: Optional[Dict] = types.MappingProxyType({})
    express_pattern = re.compile(r'\$\{(.*?)\}')


def load_config(config_dir: str):
    """仅加载`config_dir`中的application.[yml|yaml|json|properties|xml], application-dev.[yml|yaml|json|properties|xml], application-test.[yml|yaml|json|properties|xml], application-pro.[yml|yaml|json|properties|xml]
    :param config_dir: 项目目录地址
    """
    # 当前项目的配置文件
    config_file_path = _find_file_path(config_dir, '^application.(yml|yaml|json|properties|xml|py)$') or (
            config_dir + os.sep + 'application.yml')
    # 使用 active
    actives = []

    exist_config_file_path  = os.path.exists(config_file_path)

    _data_file_loader = AutoDataFileLoader()
    # 配置对象_cfg
    _cfg_dict = {}
    if exist_config_file_path:
        _cfg_dict = dict_utils.deep_update(_cfg_dict, _data_file_loader.load_file(config_file_path,
                                                                         encoding='utf-8'))
        logger.info('加载[{}]配置文件', config_file_path)

    # 优先级 env('ENV') > env('seatools.profiles.active') > application.yml('seatools.profiles.active')
    custom_env = os.getenv('ENV')
    if custom_env:
        actives.append(custom_env)
    else:
        config_actives = os.getenv ('seatools.profiles.active', ((_cfg_dict.get('seatools') or {}).get('profiles') or {}).get('active'))
        if config_actives:
            actives = [config_active.strip() for config_active in config_actives.split(',')]

    for active in actives:
        active_config_file_path = _find_file_path(config_dir, f'^application-{active}.(yml|yaml|json|properties|xml|py)$')
        if not active_config_file_path:
            continue
        if not os.path.exists(active_config_file_path):
            continue
        _cfg_dict = dict_utils.deep_update(_cfg_dict, _data_file_loader.load_file(active_config_file_path,
                                                                         encoding='utf-8'))
        logger.info('加载[{}]配置文件', active_config_file_path)

    # 解析配置中的变量
    _parse_cfg_dict(_cfg_dict, _cfg_dict)

    # 转为不可修改配置信息
    _Properties.cfg = types.MappingProxyType(_cfg_dict)


def _parse_cfg_dict(cfg_dict: dict, scan_cfg_dict: dict):
    """解析配置中的变量."""
    for k, v in scan_cfg_dict.items():
        if isinstance(v, dict):
            _parse_cfg_dict(cfg_dict, v)
            continue
        if isinstance(v, (list, tuple)):
            _parse_cfg_list(cfg_dict, v)
            continue
        scan_cfg_dict[k] = _parse_cfg_dict_value(cfg_dict, v)


def _parse_cfg_list(cfg_dict: dict, scan_cfg_list: Union[list, tuple]):
    for index in range(len(scan_cfg_list)):
        scan = scan_cfg_list[index]
        if isinstance(scan, dict):
            _parse_cfg_dict(cfg_dict, scan)
            continue
        if isinstance(scan, (list, tuple)):
            _parse_cfg_list(cfg_dict, scan)
            continue
        scan_cfg_list[index] = _parse_cfg_dict_value(cfg_dict, scan)


def _parse_cfg_dict_value(cfg_dict: dict, value):
    if not isinstance(value, str):
        return value
    expresses = list(set(_Properties.express_pattern.findall(value)))
    if not expresses:
        return value
    if len(expresses) == 1 and '${' + expresses[0] + '}' == value:
        return _parse_cfg_dict_value(cfg_dict, _parse_cfg_express_value(cfg_dict, expresses[0]))
    else:
        new_value = value
        for exp in expresses:
            new_value = new_value.replace('${' + exp + '}', f'{_parse_cfg_express_value(cfg_dict, exp)}')
        return new_value


def _parse_cfg_express_value(cfg_dict: dict, express):
    v = value_utils.parse_express_value(cfg_dict, express)
    if v is None:
        v = '${' + express + '}'
    return _parse_cfg_dict_value(cfg_dict, v)


def cfg() -> Optional[Dict]:
    """
    获取配置对象, 需要先调用 load_config, 否则返回空
    Returns:
        配置对象
    """
    return _Properties.cfg


def _find_file_path(config_dir: str, pattern: str) -> Optional[str]:
    """查找符合条件的文件路径, 如果匹配仅返回一个"""
    files = [f for f in Path(config_dir).iterdir() if f.is_file()]
    for file in files:
        if re.match(pattern, file.name):
            return config_dir + os.sep + file.name
    return None



from typing import Any, Tuple, Dict


class JoinPoint:

    def __init__(self, path, target, method_name, args: Tuple[Any, ...], kwargs: Dict[str, Any]):
        """
        Args:
           path: pointcut module-class-method path. example: xxx.xxx.xxx.XxxClass.xxx_method
           target: pointcut proxy target
           method_name: pointcut method name
           args: pointcut method args
           kwargs: pointcut method kwargs
        """
        self.target = target
        self.method_name = method_name
        self.method = getattr(target, method_name)
        self.args = args
        self.kwargs = kwargs
        self.return_value = None
        self.path = path

    def proceed(self) -> Any:
        """do point method."""
        self.return_value = self.method(*self.args, **self.kwargs)
        return self.return_value

    def get_arg(self, index = None, field_name = None):
        """Get index arg or filed_name arg. Not contains self. Note: index and field_name cannot be empty at the same time.

        Args:
            index: get index base on args
            field_name: get field name base on kwargs
        """
        if len(self.args) > index:
            return
        assert index is not None or field_name, 'index and field_name cannot be empty at the same time.'
        arg = None
        if index is not None:
            arg = self.args[index]
        if field_name:
            arg = self.kwargs.get(field_name) or arg
        return arg



from typing import Any

import pytest

from seatools.ioc import run, Autowired, Bean, Aspect, AbstractAspect
from seatools.ioc.aop.point import JoinPoint


def start():
    # 启动ioc依赖
    run(scan_package_names='tests.test_ioc_aop', config_dir='./config',
        enable_aspect=True)


@Bean
class A:

    def hello(self, name):
        print(f"hello A {name}")

    def hello_return(self, name):
        return f"hello {name}"

    def hello_error(self):
        raise RuntimeError

    def _hello_protected(self, name):
        print(f"hello A {name}")


@Aspect(order=100)
class AAop(AbstractAspect):

    pointcut = "execution(tests.test_ioc_aop.A.*)"

    def before(self, point: JoinPoint, **kwargs) -> None:
        print('A before')

    def after(self, point: JoinPoint, **kwargs) -> None:
        print('A after')

    def around(self, point: JoinPoint, **kwargs) -> Any:
        print("A around before")
        v = super().around(point, **kwargs)
        print("A around after")
        return v

    def after_returning(self, point: JoinPoint, return_value: Any, **kwargs) -> None:
        print('A after_returning')

    def after_exception(self, point: JoinPoint, ex: Exception, **kwargs) -> None:
        print('A after_exception')


class A2Aop(AbstractAspect):
    pointcut = "execution(tests.test_ioc_aop.A.*)"

    def before(self, point: JoinPoint, **kwargs) -> None:
        print('A2 before')

    def after(self, point: JoinPoint, **kwargs) -> None:
        print('A2 after')

    def around(self, point: JoinPoint, **kwargs) -> Any:
        print("A2 around before")
        v = super().around(point, **kwargs)
        print("A2 around after")
        return v

    def after_returning(self, point: JoinPoint, return_value: Any, **kwargs) -> None:
        print('A2 after_returning')

    def after_exception(self, point: JoinPoint, ex: Exception, **kwargs) -> None:
        print('A2 after_exception')


@Aspect
def a2_aop():
    return A2Aop()


def test_ioc_aop_success():
    start()
    a = Autowired(cls=A)
    a.hello("lala")
    # a._hello_protected("lalala")


def test_ioc_aop_return():
    start()
    a = Autowired(cls=A)
    print(a.hello_return("lalala"))


def test_ioc_aop_exception():
    start()
    a = Autowired(cls=A)
    with pytest.raises(RuntimeError):
        a.hello_error()




import abc
import os
import warnings

from luigi.contrib.sqla import SQLAlchemyTarget as SAT, CopyToTable as CTT
from seatools.ioc import Autowired


class SQLAlchemyTarget(SAT):
    """SQLAlchemyTarget setools.ioc implementation."""
    _engine_dict = {}

    def __init__(self, bean_name, target_table, update_id):
        super().__init__(connection_string='',
                         target_table=target_table,
                         update_id=update_id)
        self.bean_name = bean_name

    @property
    def engine(self):
        pid = os.getpid()
        conn = SQLAlchemyTarget._engine_dict.get(self.bean_name)
        if not conn or conn.pid != pid:
            # create and reset connection
            engine = Autowired(self.bean_name).engine()
            SQLAlchemyTarget._engine_dict[self.bean_name] = self.Connection(engine, pid)
        return SQLAlchemyTarget._engine_dict[self.bean_name].engine


class CopyToTable(CTT):
    """CopyToTable setools.ioc implementation."""

    @property
    def connection_string(self):
        warnings.warn('Departed. Use bean_name repeat it.', DeprecationWarning)
        return None

    @property
    @abc.abstractmethod
    def bean_name(self):
        return None

    def output(self):
        return SQLAlchemyTarget(
            bean_name=self.bean_name,
            target_table=self.table,
            update_id=self.update_id())



import pytest

from seatools.cryptography.aes import encrypt_aes, decrypt_aes
from Crypto.Random import get_random_bytes
from Crypto.Cipher import AES


def test_aes():
    # 测试用的密钥和数据
    KEY = get_random_bytes(16)  # AES-128位密钥
    IV = get_random_bytes(AES.block_size)  # 初始向量
    DATA = 'This is a top secret message.'
    data, iv = encrypt_aes(DATA.encode('utf-8'), KEY, iv=IV)
    assert data, 'AES encrypted data'
    ori_data = decrypt_aes(data, KEY, iv=iv).decode('utf-8')
    assert ori_data == DATA, 'AES decrypted data error'


def test_md5():
    from seatools.cryptography import md5, Md5DigitEnum
    text = '123456'
    assert md5(text) == 'e10adc3949ba59abbe56e057f20f883e'
    assert md5(text, digit=Md5DigitEnum.MD5_16) == '49ba59abbe56e057'
    assert md5(text, digit=Md5DigitEnum.MD5_64) == md5(text) + md5(text)
    with pytest.raises(ValueError):
        print(md5(text, digit=66))


def test_base64():
    from seatools.cryptography import encode_base16, encode_base32, encode_base64, encode_base85, encode_base32hex, \
        decode_base16, decode_base32, decode_base64, decode_base85, decode_base32hex
    text = '123456'
    assert encode_base64(text).decode('utf-8') == 'MTIzNDU2'
    assert encode_base16(text).decode('utf-8') == '313233343536'
    assert encode_base32(text).decode('utf-8') == 'GEZDGNBVGY======'
    assert encode_base32hex(text).decode('utf-8') == '64P36D1L6O======'
    assert encode_base85(text).decode('utf-8') == 'F)}kWH8u'

    assert decode_base64('MTIzNDU2').decode('utf-8') == text
    assert decode_base16('313233343536').decode('utf-8') == text
    assert decode_base32('GEZDGNBVGY======').decode('utf-8') == text
    assert decode_base32hex('64P36D1L6O======').decode('utf-8') == text
    assert decode_base85('F)}kWH8u').decode('utf-8') == text


def test_mac():
    # todo:
    pass



from seatools.ioc.database import DatabaseConfig

def test_ioc_database():
    cfg = DatabaseConfig(
        host='localhost',
        port=6379,
        password='123456',
        database=0,
        driver='redis'
    )
    print(cfg.render_to_string())
    print(cfg.render_to_string(False))



from typing import Type, TypeVar

from sqlalchemy import JSON, Dialect
from pydantic import BaseModel
from sqlalchemy.sql.type_api import _TypeMemoDict

_T = TypeVar('_T', bound=BaseModel)


class ModelJson(JSON):
    """拓展sqlalchemy json类型支持pydantic BaseModel类型"""

    __visit_name__ = "JSON"

    def __init__(self, cls: Type[_T], none_as_null: bool = False):
        super().__init__(none_as_null)
        self._cls = cls

    def bind_processor(self, dialect):
        return lambda value: self._dialect_info_impl(dialect)['impl'].bind_processor(dialect)(value.model_dump(mode='json'))


    def result_processor(self, dialect, coltype):
        return lambda value: self._cls(**(self._dialect_info_impl(dialect)['impl'].result_processor(dialect, coltype)(value)))

    def python_type(self):
        return self._cls

    def _dialect_info_impl(self, dialect: Dialect):
        return super()._dialect_info(dialect)

    def _dialect_info(self, dialect: Dialect) -> _TypeMemoDict:
        return {"impl": self, "result": {}}



import abc
import re
from typing import List, AnyStr, Optional


class AbstractMatcher(abc.ABC):

    def __init__(self, **kwargs):
        pass

    @abc.abstractmethod
    def match(self, path: str) -> bool:
        pass


class AspectPointExpressionMather(AbstractMatcher):

    def __init__(self, expression, **kwargs):
        super().__init__(**kwargs)
        self.expression = expression
        self._patterns: List[re.Pattern[AnyStr]] = [re.compile('execution\s*\((.*)\)\s*')]
        self._compile = self.compile(expression)

    def compile(self, expression: str) -> Optional[re.Pattern[AnyStr]]:
        """compile a point expression."""
        assert expression, 'expression cannot be empty.'
        if not expression:
            return None
        expression = expression.strip()
        expression = self._patterns[0].findall(expression)
        if not expression:
            return None
        expression = '^' + expression[0].strip().replace('*', '([a-zA-Z0-9][a-zA-Z0-9_]*)').replace('..', '[a-zA-Z0-9._]*') + '$'
        return re.compile(expression)

    def match(self, path: str) -> bool:
        """Check expression against a path."""
        if not self._compile:
            return False
        return True if self._compile.match(path) else False



from typing import Optional

from seatools.models import BaseModel
from seatools.ioc.database.dbconfig import DatabaseConfig


class SqlalchemyConfig(BaseModel):
    """sqlalchemy相关配置"""
    echo: Optional[bool] = False
    # 连接池回收周期
    pool_recycle: Optional[int] = 3600


class CommonDBConfig(DatabaseConfig):
    """通用 DB 配置"""
    db: Optional[str] = None
    # 是否是async连接
    is_async: Optional[bool] = False
    sqlalchemy: Optional[dict] = None



from .injects import *
from .aop import *
from .base import get_application_context, get_environment
from .boot import run
from .environment import Environment
from .context import ApplicationContext



import importlib
from typing import List, Union
from threading import RLock

from loguru import logger

from seatools.ioc.context import ApplicationContext
from seatools.ioc.utils import reflect_utils
from seatools.ioc.base import new_bean_factory, get_bean_factory
from seatools.ioc.beans.factory import BeanFactory
from seatools.ioc.config import load_config


class _Properties:
    first = True
    lock = RLock()
    loaded_modules = set()


def run(scan_package_names: Union[List[str], str], config_dir: str = None, factory: BeanFactory = None,
        exclude_modules: List[str] = None, enable_aspect: bool = False):
    """Start seatools.ioc. Allow repeated runs. A module only load once.

    Args:
        scan_package_names: Packages to scan, supports multiple packages, e.g., xxx.xxx; xxx
        config_dir: Configuration directory, if not empty, it will load application.yml and application-{env}.yml files (if exist) from the config_dir directory into the system configuration.
        It also supports injecting properties and beans through IOC using @Value and @Configuration decorators.
        factory: Bean factory, can be implemented by yourself, defaults to SimpleBeanFactory.
        exclude_modules: List of modules to exclude.
        enable_aspect: Whether to enable AOP.
    """
    # Start with locking
    with _Properties.lock:
        if not scan_package_names:
            raise ValueError('Package name cannot be empty.')
        if isinstance(scan_package_names, str):
            scan_package_names = [scan_package_names]
        # Initialize factory
        if _Properties.first:
            bean_factory = new_bean_factory(factory=factory, enable_aspect=enable_aspect)
            # Initialize configuration if provided
            if config_dir:
                load_config(config_dir=config_dir)

            # Register Environment
            from seatools.ioc.environment import Environment
            bean_factory.register_bean(name='environment', cls=Environment, primary=True, lazy=False)

            # Register ApplicationContext
            bean_factory.register_bean(name='applicationContext', cls=ApplicationContext(bean_factory), primary=True,
                                       lazy=False)

            # Add starters package for priority auto-loading
            scan_package_names = ['seatools.ioc.starters'] + scan_package_names

            _Properties.first = False
        else:
            bean_factory = get_bean_factory()

        for scan_package_name in scan_package_names:
            # Load all beans in the package and inject into the factory
            modules = reflect_utils.get_all_py_modules(scan_package_name)
            for module in modules:
                if not exclude_modules:
                    _load_module(module)
                    continue
                for exclude_module in exclude_modules:
                    if module.startswith(exclude_module):
                        break
                else:
                    _load_module(module)
        # Bean initialization
        bean_factory.init()
        # Release loading resources
        _release()


def _release():
    # Release resources
    _Properties.loaded_modules.clear()


def _load_module(module):
    if module in _Properties.loaded_modules:
        return
    # Prevent infinite recursive loading, add to record first then loa
    _Properties.loaded_modules.add(module)
    try:
        importlib.import_module(module)
    except ModuleNotFoundError as e:
        logger.warning('Module {} not found.Cannot to import by seatools.ioc. Reason: {}', module, e)



import abc
from typing import Any
from .point import JoinPoint


class AbstractAspect(abc.ABC):
    """Abstract aspect."""

    @property
    @abc.abstractmethod
    def pointcut(self) -> str:
        """Pointcut. Similar to `spring @Pointcut`

        example:
            class CustomAspect(AbstractAspect):

                pointcut = "execution(xxx.xxx..*xxx.*)" # only path and public method.

        note:
            Only support execution. Not support "annotation(...)", use decorator to repeat it.
        """
        pass

    def before(self, point: JoinPoint, **kwargs) -> None:
        """before aspect. Note: point.get_return_value() is None."""
        pass

    def around(self, point: JoinPoint, **kwargs) -> Any:
        """around aspect. Note: use point.process() to call aspect function."""
        return point.proceed()

    def after(self, point: JoinPoint, **kwargs) -> None:
        """after aspect."""
        pass

    def after_returning(self, point: JoinPoint, return_value: Any, **kwargs) -> None:
        """after success and returning aspect."""
        pass

    def after_exception(self, point: JoinPoint, ex: Exception, **kwargs) -> None:
        """after fail and raise exception aspect."""
        pass
